#include "Plane3.h"
#include "ViewCellBsp.h"
#include "Mesh.h"
#include "common.h"
#include "ViewCell.h"
#include <stack>
#include "Environment.h"
#include "Polygon3.h"

int BspTree::sTermMaxPolygons = 10;
int BspTree::sTermMaxDepth = 20;
int BspTree::sSplitPlaneStrategy = NEXT_POLYGON; 


/****************************************************************/
/*                  class BspNode implementation                */
/****************************************************************/

BspNode::BspNode(): mParent(NULL)
{}

BspNode::BspNode(BspInterior *parent): mParent(parent)
{}


bool BspNode::IsRoot() const 
{
	return mParent == NULL;
}

BspInterior *BspNode::GetParent() 
{ 
	return mParent; 
}

void BspNode::SetParent(BspInterior *parent)
{
	mParent = parent;
}

/****************************************************************/
/*              class BspInterior implementation                */
/****************************************************************/


BspInterior::BspInterior(Plane3 plane): mPlane(plane) 
{}

bool BspInterior::IsLeaf() const
{ 
	return false; 
}

BspNode *BspInterior::GetBack() 
{
	return mBack;
}

BspNode *BspInterior::GetFront() 
{
	return mFront;
}

Plane3 *BspInterior::GetPlane()
{
	return &mPlane;
}

void BspInterior::ReplaceChildLink(BspNode *oldChild, BspNode *newChild) 
{
	if (mBack == oldChild)
	{
		mBack = newChild;
	}
    else
	{
		mFront = newChild;
	}
}

void BspInterior::SetupChildLinks(BspNode *b, BspNode *f) 
{
    mBack = b;
    mFront = f;
}

void BspInterior::SplitPolygons(PolygonQueue *polys, 
								PolygonQueue *frontPolys, 
								PolygonQueue *backPolys, 
								int &splits)
{
 	while (!polys->empty())
	{
		Polygon3 *poly = polys->front();

		polys->pop();

		Debug << (*poly);

		int result = poly->Side(&mPlane);

		Polygon3 *front_piece = NULL;
		Polygon3 *back_piece = NULL;

		switch (result)
		{
			case Polygon3::COINCIDENT:
				//break; //TODO: comment in
			case Polygon3::FRONT_SIDE:
				frontPolys->push(poly);
				break;
			case Polygon3::BACK_SIDE:
				backPolys->push(poly);
				break;
			case Polygon3::SPLIT:
				front_piece = new Polygon3();
				back_piece = new Polygon3();

				//-- split polygon
				poly->Split(&mPlane, front_piece, back_piece, splits);
			
				Debug << "SPLIT, plane: " << mPlane << "\norginal: " << (*poly) << "\nback: " << (*back_piece) << "\nfront: " << (*front_piece) << endl;
				backPolys->push(back_piece);
				frontPolys->push(front_piece);

				// don't need polygon anymore
				DEL_PTR(poly);

				break;
			default:
				break;
		}
	}
}

/****************************************************************/
/*                  class BspLeaf implementation                */
/****************************************************************/

BspLeaf::BspLeaf(ViewCell *viewCell): mViewCell(viewCell) 
{
}

ViewCell *BspLeaf::GetViewCell()
{
	return mViewCell;
}

bool BspLeaf::IsLeaf() const 
{ 
	return true; 
}


/****************************************************************/
/*                  class BspTree implementation                */
/****************************************************************/

BspTree::BspTree(): mTermMaxPolygons(0), mTermMaxDepth(0)
{
	mRoot = new BspLeaf();
}
 
const BspTreeStatistics &BspTree::GetStatistics() const 
{
	return mStat;
}

void BspTreeStatistics::Print(ostream &app) const
{
	app << "===== BspTree statistics ===============\n";

	app << "#N_RAYS Number of rays )\n"
		<< rays <<endl;
	app << "#N_DOMAINS  ( Number of query domains )\n"
		<< queryDomains <<endl;

	app << "#N_NODES ( Number of nodes )\n" << nodes << "\n";

	app << "#N_LEAVES ( Number of leaves )\n" << Leaves() << "\n";

	app << "#N_SPLITS ( Number of splits)\n" << splits << "\n";

	app << "#N_RAYREFS ( Number of rayRefs )\n" <<
	rayRefs << "\n";

	app << "#N_RAYRAYREFS  ( Number of rayRefs / ray )\n" <<
	rayRefs/(double)rays << "\n";

	app << "#N_LEAFRAYREFS  ( Number of rayRefs / leaf )\n" <<
	rayRefs/(double)Leaves() << "\n";

	app << "#N_MAXOBJECTREFS  ( Max number of rayRefs / leaf )\n" <<
	maxObjectRefs << "\n";

	app << "#N_NONEMPTYRAYREFS  ( Number of rayRefs in nonEmpty leaves / non empty leaf )\n" <<
	rayRefsNonZeroQuery/(double)(Leaves() - zeroQueryNodes) << "\n";

	app << "#N_LEAFDOMAINREFS  ( Number of query domain Refs / leaf )\n" <<
	objectRefs/(double)Leaves() << "\n";

	//  app << setprecision(4);

	app << "#N_PEMPTYLEAVES  ( Percentage of leaves with zero query domains )\n"<<
	zeroQueryNodes*100/(double)Leaves()<<endl;

	app << "#N_PMAXDEPTHLEAVES ( Percentage of leaves at maxdepth )\n"<<
	maxDepthNodes * 100 / (double)Leaves() << endl;

	app << "#N_PMAXDEPTH ( Maximal reached depth )\n" << maxDepth <<endl;

	app << "#N_ADDED_RAYREFS  (Number of dynamically added ray references )\n"<<
	addedRayRefs<<endl;

	app << "#N_REMOVED_RAYREFS  (Number of dynamically removed ray references )\n"<<
	removedRayRefs<<endl;

	//  app << setprecision(4);

	//  app << "#N_CTIME  ( Construction time [s] )\n"
	//      << Time() << " \n";

	app << "===== END OF BspTree statistics ==========\n";
}

BspTree::~BspTree()
{
	std::stack<BspNode *> tStack;

	tStack.push(mRoot);

	while (!tStack.empty())
	{
		BspNode *node = tStack.top();

	    tStack.pop();
	
		if (!node->IsLeaf())
		{
			BspInterior *interior = dynamic_cast<BspInterior *>(node);

			// push the children on the stack (there are always two children)
			interior->GetBack()->mParent = NULL;
			interior->GetFront()->mParent = NULL;

			tStack.push(interior->GetBack());
			tStack.push(interior->GetFront());
		}

		DEL_PTR(node);
	}
}


void BspTree::InsertViewCell(ViewCell *viewCell)
{
	std::stack<BspTraversalData> tStack;
	
	PolygonQueue polys;
	// copy polygon information to guide the split process
	CopyMesh2Polygons(viewCell->GetMesh(), polys);

	tStack.push(BspTraversalData(mRoot, NULL, &polys, 0));

	while (!tStack.empty())
	{
		// filter polygons donw the tree
		BspTraversalData tData = tStack.top();

	    tStack.pop();
	
        if (!tData.mNode->IsLeaf())
		{
			BspInterior *interior = dynamic_cast<BspInterior *>(tData.mNode);

			// filter polygons down the tree.
			PolygonQueue *frontPolys = new PolygonQueue();
			PolygonQueue *backPolys = new PolygonQueue();

			int splits = 0;
			interior->SplitPolygons(tData.mPolygons, frontPolys, backPolys, splits);
			mStat.splits += splits;

			// push the children on the stack
			if (frontPolys->size() > 0) // if polygons on this side of bsp tree
				tStack.push(BspTraversalData(interior->GetFront(), interior->GetParent(), 
							frontPolys, tData.mDepth + 1));
			else
				delete frontPolys;

			if (backPolys > 0) // if polygons on this side of bsp tree
				tStack.push(BspTraversalData(interior->GetBack(), interior->GetParent(), 
							backPolys, tData.mDepth + 1));
			else
				delete backPolys;
		}
		else // reached leaf => subdivide
		{
			Subdivide(tStack, tData, NULL);
		}
	}
}

void BspTree::Construct(const ViewCellContainer &viewCells)
{
	// for this type of construction we split until no polygons is left
	mTermMaxPolygons = 0;
	mTermMaxDepth = sTermMaxDepth;

	mStat.nodes = 1;

	// insert all viewcells
	ViewCellContainer::const_iterator it;

	for (it = viewCells.begin(); it != viewCells.end(); ++ it)
	{
		InsertViewCell(*it);
	}
}

void BspTree::CopyMesh2Polygons(Mesh *mesh, PolygonQueue &polys)
{
	FaceContainer::const_iterator fi;
	
	// copy the face data to polygons
	for (fi = mesh->mFaces.begin(); fi != mesh->mFaces.end(); ++ fi)
	{
		Polygon3 *poly = new Polygon3((*fi), mesh);
		polys.push(poly);
	}
}

void BspTree::Copy2PolygonSoup(const ObjectContainer &objects, PolygonQueue &polys)
{
	ObjectContainer::const_iterator it, it_end = objects.end();

	int limit = min((int)objects.size(), 1);

	//for (it = objects.begin(); it != it_end; ++ it)
	for (int i = 0; i < limit; ++i)
	{
		Intersectable *object = objects[i];//*it;
		Mesh *mesh = NULL;

		// extract the meshes
		switch (object->Type())
		{
		case Intersectable::MESH_INSTANCE:
			mesh = dynamic_cast<MeshInstance *>(object)->GetMesh();
			break;
		case Intersectable::VIEWCELL:
			mesh = dynamic_cast<ViewCell *>(object)->GetMesh();
			break;
		default:
			break;
		}
		
		if (mesh) // copy the mesh data to polygons
		{
			CopyMesh2Polygons(mesh, polys);
		}
	}

	Debug << "number of polygons: " << polys.size() << endl;
}

void BspTree::Construct(const ObjectContainer &objects)
{
	Debug << "Constructing tree using object container\n";

	mTermMaxPolygons = sTermMaxPolygons;
	mTermMaxDepth = sTermMaxDepth;

	mStat.nodes = 1;

	std::stack<BspTraversalData> tStack;
	PolygonQueue *polys = new PolygonQueue();
	
	// copy mesh instance polygons into one big polygon soup
	Copy2PolygonSoup(objects, *polys);

	BspTraversalData tData(mRoot, mRoot->GetParent(), polys, 0);
	tStack.push(tData);

	while (!tStack.empty()) 
	{
		tData = tStack.top();
	    tStack.pop();
		
		Subdivide(tStack, tData);
	}
}

void BspTree::Subdivide(BspTraversalStack &tStack, BspTraversalData &tData, ViewCell *viewCell)
{
	//Debug << "Subdividing node\n";

	PolygonQueue *backPolys = new PolygonQueue();
	PolygonQueue *frontPolys = new PolygonQueue();

	BspNode *node = SubdivideNode(dynamic_cast<BspLeaf *>(tData.mNode),
								  tData.mParent,
								  tData.mPolygons,
								  tData.mDepth,
								  viewCell,
								  backPolys,
								  frontPolys);
	
	if (!node->IsLeaf()) // node was subdivided
	{
		BspInterior *interior = dynamic_cast<BspInterior *>(node);

		// push the children on the stack (there are always two children)
		tStack.push(BspTraversalData(interior->GetBack(), interior, backPolys, tData.mDepth + 1));
		tStack.push(BspTraversalData(interior->GetFront(), interior, frontPolys, tData.mDepth + 1));
	}
	else // tree terminates here
	{
		EvaluateLeafStats(tData);
		// don't need to store polygon information => delete polygons
		Polygon3::DeletePolygons(tData.mPolygons);
	}
}

Plane3 BspTree::SelectPlane(PolygonQueue *polyQueue) 
{
	// TODO: more sophisticated criteria
	switch (sSplitPlaneStrategy)
	{
	case LEAST_SPLITS:
		break;
	case NEXT_POLYGON:
        return polyQueue->front()->GetSupportingPlane();
		break;
	default:
		break;
	}
}

BspNode *BspTree::SubdivideNode(BspLeaf *leaf, 
								BspInterior *parent, 
								PolygonQueue *polys, 
								const int depth, 
								ViewCell *viewCell,
								PolygonQueue *frontPolys, 
								PolygonQueue *backPolys)
{
	// terminate traversal
	if ((polys->size() <= mTermMaxPolygons) || (depth >= mTermMaxDepth))
	{
		// Debug << polys->size() << ", " << depth << endl;		
		return leaf;
	}

	 mStat.nodes += 2;

	// add the new nodes to the tree + select subdivision plane
	BspInterior *node = new BspInterior(SelectPlane(polys)); 
	Debug << "new bspinterior: " << (*node) << endl;
	// split polygon according to current plane
	int splits = 0;
	node->SplitPolygons(polys, frontPolys, backPolys, splits);
 	mStat.splits += splits;

	// two new leaves
	BspLeaf *back = new BspLeaf();
	BspLeaf *front = new BspLeaf(viewCell);

	// replace a link from node's parent
	if (parent)
	{
		parent->ReplaceChildLink(leaf, node);
	}

	// and setup child links
	node->SetupChildLinks(back, front);

	DEL_PTR(leaf); // leaf not member of tree anymore

	return node;
}

void BspTree::ParseEnvironment()
{
	environment->GetIntValue("BspTree.Termination.maxDepth", sTermMaxDepth);
	environment->GetIntValue("BspTree.Termination.maxPolygons", sTermMaxPolygons);

	//-- extract strategy to choose the next split plane
	char splitPlaneStrategyStr[60];

	environment->GetStringValue("BspTree.splitPlaneStrategy", splitPlaneStrategyStr);
	
	sSplitPlaneStrategy = BspTree::NEXT_POLYGON;
	
	if (strcmp(splitPlaneStrategyStr, "nextPolygon") == 0)
		sSplitPlaneStrategy = BspTree::NEXT_POLYGON;
	else if (strcmp(splitPlaneStrategyStr, "leastSplits") == 0)
		sSplitPlaneStrategy = BspTree::LEAST_SPLITS;
	else 
	{
		cerr << "Wrong BSP split plane strategy " << splitPlaneStrategyStr << endl;
		exit(1);
    }
}

void BspTree::EvaluateLeafStats(const BspTraversalData &data)
{
  // the node became a leaf -> evaluate stats for leafs
  BspLeaf *leaf = dynamic_cast<BspLeaf *>(data.mNode);

  if (data.mDepth >= mTermMaxDepth)
  {
	  ++ mStat.maxDepthNodes; 
  }

  // record maximal depth
  if (data.mDepth > mStat.maxDepth)
	  mStat.maxDepth = data.mDepth;

  Debug << "BSP Traversal data. Depth: " << data.mDepth << " (max: " << mTermMaxDepth<< "), #polygons: " << 
	  data.mPolygons->size() << " (max: " << mTermMaxPolygons << ")" << endl;
}
//} // GtpVisibilityPreprocessor